from os import path

from proc import spawn
from litmus import release_taskset
from feathertrace import ftcat, SCHED_EVENTS, LOCK_EVENTS

import litmus
import topology

from model.serialize import load as load_tasks
from util.io import get_env

COMPUTE_TASK    = 'compute-task'
BACKGROUND_TASK = 'background-task'
LOCK_TASK       = 'lockstress'
SPIN_TASK       = 'spinlock-task'

RUNLENGTH = 60

EXPERIMENTS_DIR = get_env('EXPERIMENTS', '/home/bbb/dev/experiments')

def initialize_lock_file(fname):
    if not path.exists(fname):
        f = open(fname, 'w')
        # initialize to 16 zero bytes for up to 50 locks
        f.write('\x00' * 16 * 50)
        f.close()

def spinlock_task_args(task, lock_proto=None):
    args = {}
    args['P'] = int(task.period)
    args['E'] = int(task.cost)
    if not task.cpu is None:
        args['p'] = task.cpu
    if not task.priority is None:
        args['Q'] = task.priority

    if task.resmodel:
        (resid, rnum, rlen) = task.resmodel[0]
    else:
        (resid, rnum, rlen) = (0, 0, 0)

    args['r'] = rnum
    args['l'] = rnum
    args['c'] = rlen

    return '_'.join(["%s=%s" % (k, args[k]) for k in args])

def run_spin_tasks(tasks, lock_proto, wait=True):
    binary = SPIN_TASK

    args = ['-a', lock_proto,
            '-T', RUNLENGTH,
            ]

    if wait:
        args.append('-w')

    task_args = [spinlock_task_args(t, lock_proto) for t in tasks]
#    print task_args

    args += task_args

    proc = spawn(binary, args,
                 path=EXPERIMENTS_DIR)

    # wait for release
    release_taskset(num_tasks=len(tasks)).wait()

    # wait for all tasks to terminate
    proc.wait()
    assert proc.returncode == 0

def task_args(task, wait=True, lock_proto=None, irq_cpu=None):
    args = ['-P', int(task.period),
            '-E', int(task.cost),
            '-T', RUNLENGTH
            ]
    if wait:
        args.append('-w')
    if not task.cpu is None:
        args += ['-p', task.cpu]
    if not task.priority is None:
        args += ['-Q', task.priority]

    if lock_proto is None or not task.resmodel:
        binary = COMPUTE_TASK
    else:
        binary = LOCK_TASK
        (resid, rnum, rlen) = task.resmodel[0]
        lfile  =  '/tmp/resource-%s-%d.lock' % (lock_proto, resid)
        initialize_lock_file(lfile)
        args += ['-a', lock_proto]
        args += ['-r', rnum]
        args += ['-l', rnum]
        args += ['-c', rlen]
        args += ['-f', lfile]
        if irq_cpu != None and irq_cpu != 'NO_CPU':
            args += ['-i', irq_cpu]

    return (binary, args)

def run_task(task, *args, **kargs):
    (binary, args) = task_args(task, *args, **kargs)
    return spawn(binary, args,
                 path=EXPERIMENTS_DIR)

def run_tasks(tasks, lock_proto, irq_cpu):
    procs = [run_task(t, True, lock_proto, irq_cpu) for t in tasks]

    # wait for release
    release_taskset(num_tasks=len(procs)).wait()

    # wait for all tasks to terminate
    for p in procs:
        p.wait()
        assert p.returncode == 0

def trace_taskset(tasks, trace_file, lock_proto, irq_cpu):
    # for now don't do both at the same time
    events = LOCK_EVENTS if lock_proto else SCHED_EVENTS
    tracer = ftcat(trace_file, events=events)
    try:
        if lock_proto in litmus.SPINLOCKS:
            run_spin_tasks(tasks, lock_proto)
        else:
            run_tasks(tasks, lock_proto, irq_cpu)
    finally:
        # tear down ftcat
        tracer.terminate()
        tracer.wait()

def start_background_tasks(num_cpus):
    return [spawn(BACKGROUND_TASK, [cpu], path=EXPERIMENTS_DIR)
            for cpu in xrange(num_cpus)]

def stop_background_tasks(bg_tasks):
    for t in bg_tasks:
        t.terminate()
    for t in bg_tasks:
        t.wait()

def assign_cpus(tasks, sched, host):
    if sched in litmus.GLOBAL:
        for t in tasks:
            # global
            t.cpu = None
    elif sched in litmus.PARTITIONED:
        for t in tasks:
            # direct mapping
            t.cpu = int(t.partition)
    else:
        # should be a known cache level
        cache = litmus.SCHEDULERS[sched][1]
        if cache == 'ALL':
            # simulate global scheduling <=> just one cluster
            clusters = [[0]]
        else:
            clusters = topology.HOSTS[host][cache]
        for t in tasks:
            # migrate to first CPU in cluster
            t.cpu = clusters[int(t.partition)][0]
    return tasks

def assign_prios(tasks, sched):
    if sched in litmus.STATIC_PRIO:
        # assign rate-monotonic priorities
        for i,t in enumerate(sorted(tasks, key=lambda t: t.period)):
            t.priority = i + 1
    else:
        # no static priority needed
        for t in tasks:
            t.priority = None
    return tasks

def decode(name):
    params = {}
    parts = name.split('_')
    for p in parts:
        kv = p.split('=')
        k = kv[0]
        v = kv[1] if len(kv) > 1 else None
        params[k] = v
    return params

def get_config(fname):
    return path.splitext(path.basename(fname))[0]

def get_data_file(fname):
    trace_file = get_config(fname)
    return "%s.ft" % trace_file

def experiment(fname):
    tname = get_data_file(fname)

    if path.exists(tname):
        print "Skipped: %s exists." % tname
    else:
        params = decode(get_config(fname))
        sched = params['scheduler']
        host  = params['host']
        if 'locks' in params:
            sync = params['locks']
        else:
            sync = None

        tasks = load_tasks(fname)
        tasks = assign_cpus(tasks, sched, host)
        tasks = assign_prios(tasks, sched)

        bg = start_background_tasks(topology.CPUS[host])
        try:
            litmus.activate_scheduler(sched)
            irq_cpu = litmus.get_release_master()
            trace_taskset(tasks, tname, sync, irq_cpu)
            print 'Completed %s.' % fname
        finally:
            stop_background_tasks(bg)


# for each selected scheduler
# for each task set in a directory
