#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#include <sys/sysinfo.h>

#include "litmus.h"
#include "timestamp.h"

#include "busywork.h"

#define usage(fmt, args...) do { fprintf(stderr, "Error: " fmt "\n", ## args); _usage(1); }  while (0)

#define MAX_LOCKS     50
#define MAX_REQUESTS 100

#define WSS    4 /*KB*/
#define CS_WSS 1 /*KB*/

#define FEATHER_TRACE_LOG "/dev/litmus/ft_trace0"
int feather_trace_fd = -1;

pid_t mytid;

void _usage(int ret)
{
	fprintf(stderr,
		"\nUsage: lockstress [OPTIONS]\n"
		"\nOptions:\n"
		"   -f FILE\n"
		"   -r #REQUESTS-PER-JOB\n"
		"   -l NUMBER-OF-LOCKS\n"
		"   -a LOCKING-PROTOCOL (FMLP, SRP, OMLP,...)\n"
		"   -c CRITICAL-SEC-LENGTH (in ms)\n"
		"   -i DEDICATED IRQ CPU (don't use for DPCP)\n"
		"\n"
  		"   -p PARTITION\n"
		"   -E EXEC-BUDGET\n"
		"   -P PERIOD\n"
		"   -Q PRIORITY (1..%u)\n"
		"   -T TIMEOUT\n"
		"   -w\n",
		LITMUS_MAX_PRIORITY
		);
	exit(ret);
}

void ft_log(unsigned long what, cycles_t when)
{
	struct timestamp ts;
	size_t written;

	if (feather_trace_fd >= 0) {
		ts.timestamp = (uint64_t) when;
		ts.event     = (uint8_t) what;
		ts.task_type = (uint8_t) mytid;
		ts.cpu       = (uint8_t) (mytid >> 8);

		written = write(feather_trace_fd, &ts, sizeof(ts));
		if (written != sizeof(ts))
			perror("write(" FEATHER_TRACE_LOG ")");
	}
}

int open_locks(int fd, int *locks, int num_locks, int lock_type, int irq_cpu)
{
	int i;
	int config = 0;
	int num_cpus = get_nprocs();

	for (i = 0; i < num_locks; i++) {
		/* Assign resource to processor (res_id % m) under the DPCP.  */
		config = i % num_cpus;
		/* check if this CPU is unavoidable */
		if (config == irq_cpu)
			/* too bad --- use the next */
			config = (config + 1) % num_cpus;
		locks[i] = od_openx(fd, lock_type, i, &config);
		if (locks[i] < 0) {
			perror("od_open");
			return -1;
		}
	}

	return 0;
}

int lock_job(int *locks, int num_locks, int num_requests,
	     double budget, double cslen)
{
	int i;
	int err;
	volatile int sum = 0;

	double chunk = (budget - num_requests * cslen) / (1 + num_requests);

	if (chunk < 0)
		chunk = 0;

	/* initial computation */
	sum += do_busywork(WSS, chunk);

	while (num_requests--) {
		i = rand() % num_locks;

		get_ctrl_page()->ts_syscall_start = (uint64_t) get_cycles();
		err = litmus_lock(locks[i]);
		ft_log(TS_SYSCALL_OUT_END, get_cycles());
		if (err !=  0) {
			perror("litmus_lock");
			return err;
		}

		/* critical section */
		sum += do_busywork(CS_WSS, cslen);

		get_ctrl_page()->ts_syscall_start = (uint64_t) get_cycles();
		err = litmus_unlock(locks[i]);
		ft_log(TS_SYSCALL_OUT_END, get_cycles());
		if (err != 0) {
			perror("litmus_unlock");
			return err;
		}

		/* post-cs computation */
		sum += do_busywork(WSS, chunk);
	}

	sum *= 2;

	return sum;
}

#define OPTSTR "f:p:P:E:Q:T:l:r:wa:c:i:"

int main(int argc, char** argv)
{
	int opt;
	int err = 0;
	double start;

	int budget  =  25; /*ms*/
	int period  = 100; /*ms*/
	int timeout = 10;  /*s*/
	unsigned int priority = 0;
	int cpu     = 0;
	int migrate = 0;
	int wait    = 0;

	int irq_cpu = -1;

	char *fname = ".shared-locks";
	int num_requests = 3; /* per job */
	int num_locks = 1;
	double cs_length = 0.1 /*ms*/;
	int lock_type = FMLP_SEM;

	/* the locks that this task should contend */
	int fd;
	int locks[MAX_LOCKS];

	while ((opt = getopt(argc, argv, OPTSTR)) != -1) {
		switch (opt) {
		case 'l':
		        num_locks = atoi(optarg);
			if (num_locks < 1 || num_locks > MAX_LOCKS)
				usage("number of locks out of range [1,%d]", MAX_LOCKS);
			break;

		case 'r':
		        num_requests = atoi(optarg);
			if (num_requests < 1 || num_requests > MAX_REQUESTS)
				usage("number of requests out of range [1,%d]", MAX_REQUESTS);
			break;

		case 'c':
		        cs_length = atof(optarg);
			if (cs_length <= 0 || cs_length > budget)
				usage("nonsensical critical section length");
			break;

		case 'f':
			fname = optarg;
			break;

		case 'i':
			irq_cpu = atoi(optarg);
			if (irq_cpu <= 0 && strcmp(optarg, "0") != 0)
				usage("bad IRQ processor ID");
			break;

		case 'w':
			wait = 1;
			break;
		case 'p':
			cpu = atoi(optarg);
			migrate = 1;
			break;

		case 'E':
		        budget = atoi(optarg);
			break;

		case 'P':
		        period = atoi(optarg);
			break;

		case 'Q':
			priority = atoi(optarg);
			if (priority == 0 || priority > LITMUS_MAX_PRIORITY)
				usage("Invalid priority (%s).", optarg);
			break;

		case 'T':
		        timeout = atoi(optarg);
			if (timeout < 0)
				usage("invalid timeout value (%d)", timeout);
			break;

		case 'a':
			lock_type = lock_protocol_for_name(optarg);
			if (lock_type < 0)
				usage("invalid locking protocol");
			break;

		case ':':
			usage("Argument missing.");
			break;
		case '?':
		default:
			usage("Bad argument.");
			break;
		}
	}


	if (budget <= 0 || period < budget)
		usage("invalid task parameters (%d, %d)", budget, period);

	init_litmus();
	srand((int) time(NULL) + (int) getpid());
	mytid = gettid();

	err = sporadic_task(budget, period, 0, cpu, priority,
			    RT_CLASS_HARD, NO_ENFORCEMENT, migrate);
	if (err != 0) {
		fprintf(stderr, "Could not set task parameters (%m).\n");
		exit(2);
	}

	fd = open(fname, O_RDONLY | O_CREAT);
	if (fd < 0) {
		fprintf(stderr, "Could not open shared lock file (%m).\n");
		exit(3);
	}

	feather_trace_fd = open(FEATHER_TRACE_LOG, O_WRONLY | O_APPEND);
	if (feather_trace_fd < 0) {
		fprintf(stderr, "Could not open Feather-Trace log"
			"file at '%s' (%m).\n", FEATHER_TRACE_LOG);
		exit(9);
	}

	err = task_mode(LITMUS_RT_TASK);
	if (err != 0) {
		fprintf(stderr, "could not become RT task (%m)\n");
		exit(5);
	}

	err = open_locks(fd, locks, num_locks, lock_type, irq_cpu);
	if (err < 0) {
		fprintf(stderr, "Could not open locks (%m).\n");
		exit(4);
	}

	if (wait) {
		err = wait_for_ts_release();
		if (err != 0) {
			fprintf(stderr, "Could not wait for TS release (%m)\n.");
			exit(5);
		}
	}

	start = wctime();
	while (timeout == 0 || wctime() < start + timeout) {
		lock_job(locks, num_locks, num_requests, budget, cs_length);
		sleep_next_period();
	}

	/* doe one more request to trigger priority donation under the clustered OMLP */
	lock_job(locks, num_locks, 1, budget, cs_length);

	err = task_mode(BACKGROUND_TASK);
	if (err) {
		fprintf(stderr, "could not become BE task (%m)\n");
		exit(99);
	}

	/* TODO: proper resource cleanup... but the kernel does it anyway. */

	return 0;
}
