#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#include <pthread.h>


#include <sys/sysinfo.h>

/* liblitmus */
#include "litmus.h"

/* ft_tools */
#include "timestamp.h"

#include "busywork.h"

#include "atomic.h"
#include "mcs_proto.h"
#include "mcsrw_proto.h"
#include "pfq_proto.h"

#define usage(fmt, args...) do { fprintf(stderr, "Error: " fmt "\n", ## args); _usage(1); }  while (0)

#define MAX_LOCKS     50
#define MAX_REQUESTS 100
#define MAX_TASKS    500

#define WSS    4 /*KB*/
#define CS_WSS 1 /*KB*/

#define FEATHER_TRACE_LOG "/dev/litmus/ft_trace0"
int feather_trace_fd = -1;

enum  LOCK_TYPE {
	TASK_FAIR_MUTEX,
	TASK_FAIR_RW,
	PHASE_FAIR_RW,
};

union spinlock {
	uint8_t padding[CACHE_LINE_SIZE];
	mcs_lock_t    mtx;
	mcs_rwlock_t tf_rw;
	pfq_rwlock_t pf_rw;
} CACHE_LINE_ALIGNED;


union queue_node {
	mcs_node_t mtx;
	mcs_rwnode_t tf_rw;
	pfq_rwnode_t pf_rw;
} CACHE_LINE_ALIGNED;

int timeout = 10;  /*s*/
int wait = 0;
int lock_type = TASK_FAIR_MUTEX;

union spinlock locks[MAX_LOCKS];

cycles_t spin_write_lock(int idx, union queue_node* node)
{
	switch (lock_type) {
	case TASK_FAIR_MUTEX:
		return mcs_lock(&locks[idx].mtx, &node->mtx);
	case TASK_FAIR_RW:
		return mcs_write_lock(&locks[idx].tf_rw, &node->tf_rw);
	case PHASE_FAIR_RW:
		return pfq_write_lock(&locks[idx].pf_rw, &node->pf_rw);
	default:
		return 0;
	}
}

void  spin_write_unlock(int idx, union queue_node* node)
{
	switch (lock_type) {
	case TASK_FAIR_MUTEX:
		mcs_unlock(&locks[idx].mtx, &node->mtx);
		break;
	case TASK_FAIR_RW:
		mcs_write_unlock(&locks[idx].tf_rw, &node->tf_rw);
		break;
	case PHASE_FAIR_RW:
		pfq_write_unlock(&locks[idx].pf_rw, &node->pf_rw);
		break;
	}
}

cycles_t  spin_read_lock(int idx, union queue_node* node)
{
	switch (lock_type) {
	case TASK_FAIR_MUTEX:
		return mcs_lock(&locks[idx].mtx, &node->mtx);
	case TASK_FAIR_RW:
		return mcs_read_lock(&locks[idx].tf_rw, &node->tf_rw);
	case PHASE_FAIR_RW:
		return pfq_read_lock(&locks[idx].pf_rw, &node->pf_rw);
	default:
		return 0;
	}
}

void  spin_read_unlock(int idx, union queue_node* node)
{
	switch (lock_type) {
	case TASK_FAIR_MUTEX:
		mcs_unlock(&locks[idx].mtx, &node->mtx);
		break;
	case TASK_FAIR_RW:
		mcs_read_unlock(&locks[idx].tf_rw, &node->tf_rw);
		break;
	case PHASE_FAIR_RW:
		pfq_read_unlock(&locks[idx].pf_rw, &node->pf_rw);
		break;
	}
}


/* per thread:
 *  - exec
 *  - period
 *  - priority
 *  - partition
 *
 *  - requests
 *  - number of locks
 *  - critsec length
 *
 * global:
 *  - file
 *  - timeout
 *  - protocol
 */

void _usage(int ret)
{
	fprintf(stderr,
		"\nUsage: spinlock-task [OPTIONS]\n"
		"\nOptions:\n"
		"   -a LOCKING-PROTOCOL (FMLP, SRP, OMLP,...)\n"
		"   -T TIMEOUT\n"
		"   -s TASKSET\n"
		"   -w\n"
		);
	exit(ret);
}

void ft_log(unsigned long what, cycles_t when, int mytid)
{
	struct timestamp ts;
	size_t written;

	if (feather_trace_fd >= 0) {
		ts.timestamp = (uint64_t) when;
		ts.event     = (uint8_t) what;
		ts.task_type = (uint8_t) mytid;
		ts.cpu       = (uint8_t) (mytid >> 8);

		written = write(feather_trace_fd, &ts, sizeof(ts));
		if (written != sizeof(ts))
			perror("write(" FEATHER_TRACE_LOG ")");
	}
}

int lock_job(int num_locks, int num_requests,
	     double budget, double cslen, int mytid, struct drand48_data* prng)
{
	int i;
	int err;
	long num;
	volatile int sum = 0;
	cycles_t start, stop, in, out, spin;

	union queue_node qnode;

	double chunk = (budget - num_requests * cslen) / (1 + (num_requests * 2));

	if (chunk < 0)
		chunk = 0;

	/* initial computation */
	sum += do_busywork(WSS, chunk);

	while (num_requests--) {
		lrand48_r(prng, &num);
		i = num % num_locks;

		/* READ-SIDE CS */
		start = get_cycles();
		enter_np();
		/* spin lock */
		spin = spin_read_lock(i, &qnode);
		in  = get_cycles();

		/* critical section */
		sum += do_busywork(CS_WSS, cslen);


		/* spin unlock */
		out = get_cycles();
		spin_read_unlock(i, &qnode);
		stop = get_cycles();
		if (exit_np_trace())
			ft_log(TS_SYSCALL_OUT_END, get_cycles(), mytid);

		ft_log(TS_READ_LOCK_START, start, mytid);
		/* take out time spent spinning */
		ft_log(TS_READ_LOCK_END, in - spin, mytid);
		ft_log(TS_READ_UNLOCK_START, out, mytid);
		ft_log(TS_READ_UNLOCK_END, stop, mytid);


//		printf("[%d] r-spin=%llu\n", mytid, (unsigned long long) spin);

		/* WRITE-SIDE CS */
		start = get_cycles();
		enter_np();
		/* spin lock */
		spin = spin_write_lock(i, &qnode);
		in  = get_cycles();

		/* critical section */
		sum += do_busywork(CS_WSS, cslen);


		/* spin unlock */
		out = get_cycles();
		spin_write_unlock(i, &qnode);
		stop = get_cycles();
		if (exit_np_trace())
			ft_log(TS_SYSCALL_OUT_END, get_cycles(), mytid);

		ft_log(TS_LOCK_START, start, mytid);
		/* take out time spent spinning */
		ft_log(TS_LOCK_END, in - spin, mytid);
		ft_log(TS_UNLOCK_START, out, mytid);
		ft_log(TS_UNLOCK_END, stop, mytid);

//		printf("[%d] w-spin=%llu\n", mytid, (unsigned long long) spin);

		/* post-cs computation */
		sum += do_busywork(WSS, chunk);
	}

	sum *= 2;

	err = 0;

	return sum;
}

int kv_get_param(const char* param_string, const char* key,
		 char* buf, size_t maxlen)
{
	char *pos, *end;
	size_t n;

	pos = strstr(param_string, key);

	if (!pos) {
		return 0;
	}

	end = strchr(pos, '_');
	pos = strchr(pos, '=');

	if (!pos || (end && pos > end)) {
		/* no parameter present */
		strncpy(buf, "", maxlen);
	} else {
		pos++;
		/* pos < end =>  parameter present */
		if (!end)
			end = strchr(pos, '\0');
		n = end - pos > maxlen - 1 ? maxlen - 1 : end - pos;
		strncpy(buf, pos, n);
		buf[n] = '\0';
	}

	return 1;
}

void* spinlock_task(void* arg)
{
	char* params = (char*) arg;

	int mytid;
	int err = 0;
	double start;

	int budget  =  25; /*ms*/
	int period  = 100; /*ms*/
	unsigned int priority = 0;
	int cpu     = 0;
	int migrate = 0;

	int num_requests = 3; /* per job */
	int num_locks = 1;
	double cs_length = 0.1 /*ms*/;

	struct drand48_data prng;

	char buf[64];
#define buflen sizeof(buf)

	mytid = gettid();
	srand48_r(time(NULL) + mytid, &prng);

/*
	printf("spinlock_task/%d %s\n", mytid, params);
*/

	/* parse params */

	if (kv_get_param(params, "l", buf, buflen)) {
		num_locks = atoi(buf);
		if (num_locks < 1 || num_locks > MAX_LOCKS)
			usage("number of locks out of range [1,%d] [%s]", MAX_LOCKS, params);
	}

	if (kv_get_param(params, "r", buf, buflen)) {
		num_requests = atoi(buf);
		if (num_requests < 1 || num_requests > MAX_REQUESTS)
			usage("number of requests out of range [1,%d] [%s]", MAX_REQUESTS, params);
	}

	if (kv_get_param(params, "c", buf, buflen)) {
		cs_length = atof(buf);
		if (cs_length <= 0 || cs_length > budget)
			usage("nonsensical critical section length [%s]", params);
	}

	if (kv_get_param(params, "p", buf, buflen)) {
		cpu = atoi(buf);
		migrate = 1;
	}

	if (kv_get_param(params, "E", buf, buflen)) {
		budget = atoi(buf);
	}

	if (kv_get_param(params, "P", buf, buflen)) {
		period = atoi(buf);
	}

	if (kv_get_param(params, "Q", buf, buflen)) {
		priority = atoi(buf);
		if (priority == 0 || priority > LITMUS_MAX_PRIORITY)
			usage("Invalid priority (%s).", buf);
	}

	if (budget <= 0 || period < budget)
		usage("invalid task parameters (%d, %d)", budget, period);


	init_litmus();

	err = sporadic_task(budget, period, 0, cpu, priority,
			    RT_CLASS_HARD, NO_ENFORCEMENT, migrate);
	if (err != 0) {
		fprintf(stderr, "Could not set task parameters (%m). [%s]\n", params);
		exit(2);
	}


	err = task_mode(LITMUS_RT_TASK);
	if (err != 0) {
		fprintf(stderr, "could not become RT task (%m) [%s]\n", params);
		exit(5);
	}

	if (wait) {
		err = wait_for_ts_release();
		if (err != 0) {
			fprintf(stderr, "Could not wait for TS release (%m)\n.");
			exit(5);
		}
	}

	start = wctime();
	while (timeout == 0 || wctime() < start + timeout) {
		lock_job(num_locks, num_requests, budget, cs_length, mytid, &prng);
		sleep_next_period();
	}

	return NULL;
}


#define OPTSTR "wT:a:"

int main(int argc, char** argv)
{
	int opt;
	int err = 0;
	int task_count = 0, i;

	pthread_t tasks[MAX_TASKS];

	BUILD_BUG_ON(sizeof(union spinlock) != CACHE_LINE_SIZE);

	while ((opt = getopt(argc, argv, OPTSTR)) != -1) {
		switch (opt) {
		case 'w':
			wait = 1;
			break;

		case 'T':
		        timeout = atoi(optarg);
			if (timeout < 0)
				usage("invalid timeout value (%d)", timeout);
			break;

		case 'a':
			if (strcmp(optarg, "MX-Q") == 0)
				lock_type = TASK_FAIR_MUTEX;
			else if (strcmp(optarg, "TF-Q") == 0)
				lock_type = TASK_FAIR_RW;
			else if (strcmp(optarg, "PF-Q") == 0)
				lock_type = PHASE_FAIR_RW;
			else
				usage("invalid locking protocol");
			break;

		case ':':
			usage("Argument missing.");
			break;
		case '?':
		default:
			usage("Bad argument.");
			break;
		}
	}

	feather_trace_fd = open(FEATHER_TRACE_LOG, O_WRONLY | O_APPEND);
	if (feather_trace_fd < 0) {
		fprintf(stderr, "Could not open Feather-Trace log"
			"file at '%s' (%m).\n", FEATHER_TRACE_LOG);
		exit(9);
	}

	while (optind < argc) {
		if (task_count >= MAX_TASKS) {
			fprintf(stderr, "too many tasks\n");
			exit(11);
		}

		err = pthread_create(&tasks[task_count], NULL, spinlock_task, argv[optind]);
		if (err) {
			fprintf(stderr, "could not launch spinlock_task (%m)\n");
			exit(12);
		}
		task_count++;
		optind++;
	}

	for (i = 0; i < task_count; i++) {
		err = pthread_join(tasks[i], NULL);
		if (err) {
			fprintf(stderr, "could not join (%s) (%m)\n", strerror(err));
			exit(14);
		}
	}

	/* TODO: proper resource cleanup... but the kernel does it anyway. */

	return 0;
}
